#include <mysql/mysql.h>
#include <sstream>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <cstring>

static const auto RET_SUCCESS = 0;
static const auto RET_FAIL = 1;

static void gMysqlError(const char* apFuncName, MYSQL* apMysql)
{
	std::ostringstream vStream;
	vStream
		<< "Error: "
		<< apFuncName << " "
		<< mysql_sqlstate(apMysql) << ","
		<< mysql_errno(apMysql) << ","
		<< mysql_error(apMysql);
	throw std::runtime_error {vStream.str()};
}

static void gMysqlStmtError(const char* apFuncName, MYSQL_STMT* apMysqlStmt)
{
	std::ostringstream vStream;
	vStream
		<< "Error: "
		<< apFuncName << " "
		<< mysql_stmt_sqlstate(apMysqlStmt) << ","
		<< mysql_stmt_errno(apMysqlStmt) << ","
		<< mysql_stmt_error(apMysqlStmt);
	mysql_stmt_close(apMysqlStmt);
	throw std::runtime_error {vStream.str()};
}

static void gGetOutParams(MYSQL* apMysql, MYSQL_STMT* apMysqlStmt)
{
	// Obtain output parameter metadata
	auto pMysqlRes = mysql_stmt_result_metadata(apMysqlStmt);
	if (pMysqlRes == nullptr)
	{
		gMysqlStmtError("mysql_stmt_result_metadata()", apMysqlStmt);
	}
	auto pMysqlField = mysql_fetch_field_direct(pMysqlRes, 0);
	std::cout << "Output parameter name = " << pMysqlField->name << std::endl;
	std::cout << "Output parameter type = " << pMysqlField->type << std::endl;
	std::cout << "Output parameter length = " << pMysqlField->length << std::endl;
	mysql_free_result(pMysqlRes);

	// Bind output parameter
	char vOutVal;
	unsigned long vOutValLen;
	my_bool vOutValNull;
	MYSQL_BIND vOutBind;
	vOutBind.buffer_type = MYSQL_TYPE_STRING;
	vOutBind.buffer = &vOutVal;
	vOutBind.buffer_length = sizeof(vOutVal);
	vOutBind.length = &vOutValLen;
	vOutBind.is_null = &vOutValNull;
	vOutBind.is_unsigned = 0;
	vOutBind.error = nullptr;
	if (mysql_stmt_bind_result(apMysqlStmt, &vOutBind) != 0)
	{
		gMysqlStmtError("mysql_stmt_bind_result()", apMysqlStmt);
	}

	// Obtain output parameter value
	auto vResult = mysql_stmt_fetch(apMysqlStmt);
	if (vResult == 1)
	{
		gMysqlStmtError("mysql_stmt_fetch()", apMysqlStmt);
	}
	std::cout << "Output parameter value = " << vOutVal << std::endl;
	mysql_stmt_free_result(apMysqlStmt);
}

int main()
{
	try
	{
		// CEnvironment
		if (mysql_library_init(0, nullptr, nullptr))
		{
			std::ostringstream vStream;
			vStream << "Error: mysql_library_init() failed";
			throw std::runtime_error {vStream.str()};
		}

		// CConnection
		MYSQL* pMysql = nullptr;
		if ((pMysql = mysql_init(pMysql)) == nullptr)
		{
			std::ostringstream vStream;
			vStream << "Error: mysql_init() failed";
			throw std::runtime_error {vStream.str()};
		}
		mysql_options(pMysql, MYSQL_READ_DEFAULT_FILE, "config/my.cnf");
		if
		(
			mysql_real_connect
			(
				pMysql,
				"192.168.0.131",
				"user",
				"password",
				"testdb",
				0,
				nullptr,
				CLIENT_MULTI_RESULTS
			) == nullptr
		)
		{
			gMysqlError("mysql_real_connect()", pMysql);
		}
		if (mysql_autocommit(pMysql, 1))
		{
			gMysqlError("mysql_autocommit()", pMysql);
		}

		// CStoredProcedure
		// - Create statement
		MYSQL_STMT* pMysqlStmt;
		if ((pMysqlStmt = mysql_stmt_init(pMysql)) == nullptr)
		{
			gMysqlError("mysql_stmt_init()", pMysql);
		}
		std::string vStmtText = "call test_in_out_1(?, ?)";
		if (mysql_stmt_prepare(pMysqlStmt, vStmtText.c_str(), vStmtText.length()) != 0)
		{
			gMysqlStmtError("mysql_stmt_prepare()", pMysqlStmt);
		}
		auto vNumParams = mysql_stmt_param_count(pMysqlStmt);
		std::cout << "Number of parameters = " << vNumParams << std::endl;
		// - Bind parameters
		char vInVal[2] = { 'N', 'N' };
		unsigned long vInValLen[2] = { sizeof(vInVal[0]), sizeof(vInVal[1]) };
		my_bool vInValNull[2] = { 0, 0 };
		MYSQL_BIND vInBind[2];
		std::memset(vInBind, 0, sizeof(vInBind));
		vInBind[0].buffer_type = MYSQL_TYPE_STRING;
		vInBind[0].buffer = &vInVal[0];
		vInBind[0].buffer_length = sizeof(vInVal[0]);
		vInBind[0].length = &vInValLen[0];
		vInBind[0].is_null = &vInValNull[0];
		vInBind[0].is_unsigned = 0;
		vInBind[0].error = nullptr;
		vInBind[1].buffer_type = MYSQL_TYPE_STRING;
		vInBind[1].buffer = &vInVal[1];
		vInBind[1].buffer_length = sizeof(vInVal[1]);
		vInBind[1].length = &vInValLen[1];
		vInBind[1].is_null = &vInValNull[1];
		vInBind[1].is_unsigned = 0;
		vInBind[1].error = nullptr;
		if (mysql_stmt_bind_param(pMysqlStmt, vInBind) != 0)
		{
			gMysqlStmtError("mysql_stmt_bind_param()", pMysqlStmt);
		}
		// - Execute statement
		if (mysql_stmt_execute(pMysqlStmt) != 0)
		{
			gMysqlStmtError("mysql_stmt_execute()", pMysqlStmt);
		}
		// - Check for results
		int vNext = 0;
		do
		{
			auto vResultSetCount = mysql_stmt_field_count(pMysqlStmt);
			if (vResultSetCount != 0 && pMysql->server_status & SERVER_PS_OUT_PARAMS)
			{
				std::cout << "Number of output parameters detected = " << vResultSetCount << std::endl;
				gGetOutParams(pMysql, pMysqlStmt);
			}
			vNext = mysql_stmt_next_result(pMysqlStmt);
			if (vNext > 0)
			{
				gMysqlStmtError("mysql_stmt_next_result()", pMysqlStmt);
			}
		} while (vNext == 0);

		// ~CStoredProcedure
		mysql_stmt_close(pMysqlStmt);

		// ~CConnection
		mysql_close(pMysql);

		// ~CEnvironment
		mysql_library_end();

		return RET_SUCCESS;
	}
	catch (std::exception& crException)
	{
		std::cerr << crException.what() << std::endl;
		return RET_FAIL;
	}
	catch (...)
	{
		std::cerr << "Unknown exception" << std::endl;
		return RET_FAIL;
	}
}
